@@ -12,6 +12,7 @@ use std::os::unix::prelude::FileExt as UnixFileExt;
 use std::path::Path;
 
 type PageId = u32; // address of page in the file
+type PageState = u16; // page state bit mask
 type BufferId = u32; // index of page in buffer cache
                      // (Sub)transaction identifier: monotonically increased counter of subtransactions
                      // This identifier is used to detect deteriorate state of iterator and maintaining COW map for subtransactions.
@@ -389,7 +390,7 @@ struct PageHeader {
     list: L2List,          // LRU l2-list
     dirty_index: BufferId, // index in dirty_pages vector
     access_count: u16,
-    state: u16, // bitmask of PAGE_RAW, PAGE_DIRTY, ...
+    state: PageState, // bitmask of PAGE_RAW, PAGE_DIRTY, ...
 }
 
 impl PageHeader {
@@ -449,10 +450,12 @@ struct Database {
     alloc_page_pos: usize,              // allocator current position in internal page
     cow_map: HashMap<PageId, CowEntry>, // copy-on-write map: we have to store this mapping to avoid creation of redundant copies
     pending_deletes: Vec<PageId>, // we can no immediately delete pages, because deleted page can be reused and can not be stored in `cow` map
+    overwritten_pages: Vec<PageId>, // overwritten page copies created by subtransactions
     n_alloc_pages: u64,           // number of allocated pages since session start
     n_free_pages: u64,            // number of deallocated pages since session start
-    n_committed_txns: u64,        // number of committed transactions
+    n_committed_txns: u64,        // number of committed transactions[
     n_aborted_txns: u64,          // number of aborted transactions
+    n_overwritten_pages: u64,     // number of copies overwritten by subtransactions
 }
 
 impl Database {
@@ -491,6 +494,7 @@ struct BufferManager {
 
     hash_table: Vec<BufferId>, // array containing indexes of collision chains
     buffers: Vec<PageHeader>,  // page data
+    overwritten: Vec<PageId>,  // vector of buffers with page copies overwritten in subtransactions
 }
 
 //
@@ -744,6 +748,18 @@ impl PageData {
 }
 
 impl BufferManager {
+    fn get_buffer_pid(&self, id: BufferId) -> PageId {
+        self.buffers[id as usize].pid
+    }
+
+    fn get_buffer_state(&self, id: BufferId) -> PageState {
+        self.buffers[id as usize].state
+    }
+
+    fn set_buffer_state(&mut self, id: BufferId, state: PageState) {
+        self.buffers[id as usize].state = state
+    }
+
     //
     // Link buffer to the head of LRU list (make it acceptable for eviction)
     //
@@ -826,6 +842,7 @@ impl BufferManager {
         let mut buf = self.hash_table[hash];
         while buf != 0 {
             if self.buffers[buf as usize].pid == pid {
+                self.undirty_buffer(buf);
                 self.throw_buffer(buf);
                 break;
             }
@@ -921,12 +938,13 @@ impl BufferManager {
                 let victim = self.lru.prev;
                 ensure!(victim != 0);
                 debug_assert!(self.buffers[victim as usize].access_count == 0);
-                if (self.buffers[victim as usize].state & PAGE_DIRTY) != 0 {
+                let state = self.buffers[victim as usize].state;
+                if (state & PAGE_DIRTY) != 0 {
                     // remove page from dirty vector
                     self.undirty_buffer(victim);
                     evicted_pid = self.buffers[victim as usize].pid;
                 }
-                debug_assert!((self.buffers[victim as usize].state & (PAGE_BUSY | PAGE_RAW)) == 0);
+                debug_assert!((state & (PAGE_BUSY | PAGE_RAW)) == 0);
                 self.pin(victim);
                 self.remove(victim);
                 buf = victim;
@@ -980,56 +998,82 @@ impl<'a> PageGuard<'a> {
     fn modify(&mut self, db: &mut Database, bitmap_page: Option<usize>) -> Result<()> {
         let mut bm = self.storage.buf_mgr.lock();
         let old_buf = self.buf;
-        let old_pid = bm.buffers[old_buf as usize].pid;
+        let old_pid = bm.get_buffer_pid(old_buf);
         let just_modified = bm.modify_buffer(old_buf);
-        drop(bm);
 
         // Check if there is not mapping for this page in this subtransaction,
         // Allocator bitmap pages are not accessed by read-only queries, so there is no need to clone
         // bitmap pages for subtrasactions.
-        if db.cow_map.get(&old_pid).map_or(just_modified, |entry| {
-            entry.xid != db.meta.xid && !bitmap_page.is_some()
-        }) {
-            // Create copy
-            let new_pid = self.storage.allocate_page(db)?;
-            db.cow_map.insert(
-                new_pid,
-                CowEntry {
-                    pid: old_pid,
-                    xid: db.meta.xid,
-                },
-            ); // remember COW mapping
-            self.pid = new_pid;
-            let mgr = &self.storage.buf_mgr;
-            let (new_buf, evicted_pid) =
-                mgr.lock().copy_on_write(old_buf, new_pid, just_modified)?;
-            if evicted_pid != 0 {
-                // dirty page was evicted
-                let page = self.storage.pool[new_buf as usize].read();
-                self.storage
-                    .file
-                    .write_all_at(&page.data, evicted_pid as u64 * PAGE_SIZE as u64)?;
-            }
-            if new_buf != old_buf {
-                // Copy page content to the new buffer
-                let mut dst = self.storage.pool[new_buf as usize].write();
-                let src = self.storage.pool[old_buf as usize].read();
-                *dst = *src;
-                let mut bm = mgr.lock();
-                debug_assert!(bm.buffers[new_buf as usize].state == PAGE_RAW);
-                if just_modified {
-                    bm.undirty_buffer(old_buf);
+        let orig_pid: PageId;
+        match db.cow_map.get(&old_pid) {
+            None => {
+                if !just_modified {
+                    // page already marked as dirty (recusion inside allocator)
+                    return Ok(());
                 }
-                bm.release_buffer(old_buf);
-                assert!(bm.modify_buffer(new_buf));
-                self.buf = new_buf;
+                // page was not yet updated by this transactions
+                orig_pid = old_pid;
             }
-            // Prevent infinite recursion
-            if let Some(alloc_page) = bitmap_page {
-                db.meta.alloc_pages[alloc_page] = new_pid;
+            Some(entry) => {
+                // Page was created or updated by this transactions
+                if entry.xid == db.meta.xid || bitmap_page.is_some() {
+                    // Page is updated by this subtransaction or is bitmap page:
+                    // in this case we do not need to create new copy
+                    return Ok(());
+                }
+                // Overwitten page was already copied by some subtransaction.
+                // We still have to create copy to provide isolation.
+                // But this copy can be removed once this subtransaction is completed.
+                bm.overwritten.push(old_pid);
+                if entry.pid != 0 {
+                    db.pending_deletes.push(entry.pid);
+                }
+                db.cow_map.remove(&old_pid).unwrap();
+                orig_pid = 0;
             }
-            self.storage.prepare_free_page(db, old_pid)?; // prepare deallocation of original page
         }
+        drop(bm);
+
+        // Create copy
+        let new_pid = self.storage.allocate_page(db)?;
+        db.cow_map.insert(
+            new_pid,
+            CowEntry {
+                pid: orig_pid,
+                xid: db.meta.xid,
+            },
+        ); // remember COW mapping
+
+        self.pid = new_pid;
+        let mut bm = self.storage.buf_mgr.lock();
+        let (new_buf, evicted_pid) = bm.copy_on_write(old_buf, new_pid, just_modified)?;
+        if evicted_pid != 0 {
+            // dirty page was evicted
+            let page = self.storage.pool[new_buf as usize].read();
+            self.storage
+                .file
+                .write_all_at(&page.data, evicted_pid as u64 * PAGE_SIZE as u64)?;
+        }
+        if new_buf != old_buf {
+            // Copy page content to the new buffer
+            let mut dst = self.storage.pool[new_buf as usize].write();
+            let src = self.storage.pool[old_buf as usize].read();
+            *dst = *src;
+            debug_assert!(bm.get_buffer_state(new_buf) == PAGE_RAW);
+            if just_modified {
+                bm.undirty_buffer(old_buf);
+            }
+            bm.release_buffer(old_buf);
+            assert!(bm.modify_buffer(new_buf));
+            self.buf = new_buf;
+        }
+        drop(bm);
+
+        // Prevent infinite recursion
+        if let Some(alloc_page) = bitmap_page {
+            db.meta.alloc_pages[alloc_page] = new_pid;
+        }
+        self.storage.prepare_free_page(db, old_pid)?; // prepare deallocation of original page
         Ok(())
     }
 }
@@ -1044,10 +1088,29 @@ impl<'a> Drop for PageGuard<'a> {
 // Storage internal methods implementations
 //
 impl Storage {
+    //
+    // Remove copy-on-write pages overwritten by completed subtransaction.
+    //
+    fn remove_redundand_copies(&self, db: &mut Database) -> Result<()> {
+        let mut overwritten: Vec<PageId> = Vec::new();
+        let mut bm = self.buf_mgr.lock();
+        std::mem::swap(&mut bm.overwritten, &mut overwritten);
+        db.n_overwritten_pages += overwritten.len() as u64;
+        for pid in &overwritten {
+            bm.forget_buffer(*pid);
+        }
+        db.overwritten_pages.append(&mut overwritten);
+        Ok(())
+    }
+
     //
     // Allocate page
     //
     fn allocate_page(&self, db: &mut Database) -> Result<PageId> {
+        // Fast path: try to reuse overwritten page
+        if let Some(pid) = db.overwritten_pages.pop() {
+            return Ok(pid);
+        }
         db.n_alloc_pages += 1;
         //
         // Cyclic scan of bitmap pages. We start with position pointed by (alloc_page,alloc_page_pos)
@@ -1139,9 +1202,14 @@ impl Storage {
                 assert!(self.pool[pin.buf as usize].write().test_and_set_bit(bit))
             }
             BitmapOp::Clear => {
-                // Set allocator's current position to just deallocated page to force its reusing
-                db.alloc_page = alloc_page;
-                db.alloc_page_pos = bit / 8;
+                // Set allocator's current position if deallocated page is before it
+                let alloc_page_pos = bit / 8;
+                if alloc_page < db.alloc_page
+                    || (alloc_page == db.alloc_page && alloc_page_pos < db.alloc_page_pos)
+                {
+                    db.alloc_page = alloc_page;
+                    db.alloc_page_pos = alloc_page_pos;
+                }
                 self.pool[pin.buf as usize].write().clear_bit(bit)
             }
         }
@@ -1224,7 +1292,7 @@ impl Storage {
     fn get_page(&self, pid: PageId) -> Result<PageGuard<'_>> {
         let mut bm = self.buf_mgr.lock();
         let (buf, evicted_pid) = bm.get_buffer(pid)?;
-        let state = bm.buffers[buf as usize].state;
+        let state = bm.get_buffer_state(buf);
         if evicted_pid != 0 {
             // dirty page was evicted
             let page = self.pool[buf as usize].read();
@@ -1233,18 +1301,18 @@ impl Storage {
         }
         if (state & PAGE_BUSY) != 0 {
             // Some other thread is loading buffer: just wait until it done
-            bm.buffers[buf as usize].state |= PAGE_WAIT;
+            bm.set_buffer_state(buf, state | PAGE_WAIT);
             loop {
-                debug_assert!((bm.buffers[buf as usize].state & PAGE_WAIT) != 0);
+                debug_assert!((bm.get_buffer_state(buf) & PAGE_WAIT) != 0);
                 self.busy_events[buf as usize % N_BUSY_EVENTS].wait(&mut bm);
-                if (bm.buffers[buf as usize].state & PAGE_BUSY) == 0 {
+                if (bm.get_buffer_state(buf) & PAGE_BUSY) == 0 {
                     break;
                 }
             }
-        } else if (bm.buffers[buf as usize].state & PAGE_RAW) != 0 {
+        } else if (state & PAGE_RAW) != 0 {
             // Read page
-            debug_assert!(bm.buffers[buf as usize].state == PAGE_RAW);
-            bm.buffers[buf as usize].state = PAGE_BUSY;
+            debug_assert!(state == PAGE_RAW);
+            bm.set_buffer_state(buf, PAGE_BUSY);
             drop(bm); // read page without holding lock
             {
                 let mut page = self.pool[buf as usize].write();
@@ -1252,12 +1320,12 @@ impl Storage {
                     .read_exact_at(&mut page.data, pid as u64 * PAGE_SIZE as u64)?;
             }
             bm = self.buf_mgr.lock();
-            if (bm.buffers[buf as usize].state & PAGE_WAIT) != 0 {
+            if (bm.get_buffer_state(buf) & PAGE_WAIT) != 0 {
                 // Somebody is waiting for us
                 self.busy_events[buf as usize % N_BUSY_EVENTS].notify_all();
             }
-            debug_assert!((bm.buffers[buf as usize].state & !(PAGE_BUSY | PAGE_WAIT)) == 0);
-            bm.buffers[buf as usize].state = 0;
+            debug_assert!((bm.get_buffer_state(buf) & !(PAGE_BUSY | PAGE_WAIT)) == 0);
+            bm.set_buffer_state(buf, 0);
         }
         Ok(PageGuard {
             buf,
@@ -1318,11 +1386,19 @@ impl Storage {
             for pg in &pending_deletes {
                 self.prepare_free_page(db, *pg)?;
             }
+            let mut overwritten_pages: Vec<PageId> = Vec::new();
+            std::mem::swap(&mut db.overwritten_pages, &mut overwritten_pages);
+            for pg in &overwritten_pages {
+                self.prepare_free_page(db, *pg)?;
+            }
             // ... then actually do it
             let n_alloc_pages = db.n_alloc_pages;
             for pg in &pending_deletes {
                 self.free_page(db, *pg)?;
             }
+            for pg in &overwritten_pages {
+                self.free_page(db, *pg)?;
+            }
             // Delete original pages (them should be already prepared)
             let orig: Vec<PageId> = db
                 .cow_map
@@ -1346,7 +1422,6 @@ impl Storage {
             if !self.conf.nosync {
                 self.file.sync_all()?;
             }
-
             db.n_committed_txns += 1;
 
             let mut meta_shadow = self.meta_shadow.write();
@@ -1383,7 +1458,7 @@ impl Storage {
             // Just throw away all dirty buffers from buffer cache to force reloading of original pages
             for i in 0..bm.dirty_buffers.len() {
                 let buf = bm.dirty_buffers[i];
-                debug_assert!((bm.buffers[buf as usize].state & PAGE_DIRTY) != 0);
+                debug_assert!((bm.get_buffer_state(buf) & PAGE_DIRTY) != 0);
                 bm.throw_buffer(buf);
             }
             bm.dirty_buffers.clear();
@@ -1405,6 +1480,7 @@ impl Storage {
 
             db.n_aborted_txns += 1;
             db.pending_deletes.clear();
+            db.overwritten_pages.clear();
         }
     }
 
@@ -1457,6 +1533,7 @@ impl Storage {
                 pinned: 1,
                 hash_table: vec![0; conf.cache_size],
                 buffers: vec![PageHeader::new(); conf.cache_size],
+                overwritten: Vec::new(),
             }),
             pool: iter::repeat_with(|| RwLock::new(PageData::new()))
                 .take(conf.cache_size)
@@ -1473,10 +1550,12 @@ impl Storage {
                 alloc_page_pos: 0,
                 cow_map: HashMap::new(),
                 pending_deletes: Vec::new(),
+                overwritten_pages: Vec::new(),
                 n_alloc_pages: 0,
                 n_free_pages: 0,
                 n_committed_txns: 0,
                 n_aborted_txns: 0,
+                n_overwritten_pages: 0,
             }),
         };
         Ok(storage)
@@ -2174,6 +2253,7 @@ impl<'a> Transaction<'_> {
         self.db.meta.xid = self.db.meta.xid.wrapping_add(1);
         let mut meta = self.storage.meta_shadow.write();
         meta.snapshot.update(&self.db.meta);
+        self.storage.remove_redundand_copies(&mut self.db)?;
         // mark transaction as committed to prevent implicit rollback by destructor
         self.status = TransactionStatus::Committed;
         Ok(())
